@@ -1,10 +1,12 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import stdnum.eu.vat as vat
+from stdnum import get_cc_module
 import stdnum.exceptions
 from sql import Null, Column, Literal
+from sql.aggregate import Min
 from sql.functions import CharLength, Substring, Position
 
+from trytond.i18n import gettext
 from trytond.model import (ModelView, ModelSQL, MultiValueMixin, ValueMixin,
     DeactivableMixin, fields, Unique, sequence_ordered)
 from trytond.wizard import Wizard, StateTransition, StateView, Button
@@ -13,24 +15,17 @@ from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond import backend
 from trytond.tools.multivalue import migrate_property
-
-__all__ = ['Party', 'PartyLang', 'PartyCategory', 'PartyIdentifier',
-    'CheckVIESResult', 'CheckVIES',
-    'PartyReplace', 'PartyReplaceAsk',
-    'PartyErase', 'PartyEraseAsk']
-
-VAT_COUNTRIES = [('', '')]
-STATES = {
-    'readonly': ~Eval('active', True),
-}
-DEPENDS = ['active']
+from trytond.tools import lstrip_wildcard
+from .exceptions import (
+    InvalidIdentifierCode, VIESUnavailable, SimilarityWarning, EraseError)
 
 
 class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     "Party"
     __name__ = 'party.party'
 
-    name = fields.Char('Name', select=True, states=STATES, depends=DEPENDS,
+    name = fields.Char(
+        "Name", select=True,
         help="The main identifier of the party.")
     code = fields.Char('Code', required=True, select=True,
         states={
@@ -41,23 +36,22 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     code_readonly = fields.Function(fields.Boolean('Code Readonly'),
         'get_code_readonly')
     lang = fields.MultiValue(
-        fields.Many2One('ir.lang', "Language", states=STATES, depends=DEPENDS,
+        fields.Many2One('ir.lang', "Language",
             help="Used to translate communications with the party."))
     langs = fields.One2Many(
         'party.party.lang', 'party', "Languages")
-    identifiers = fields.One2Many('party.identifier', 'party', 'Identifiers',
-        states=STATES, depends=DEPENDS,
+    identifiers = fields.One2Many(
+        'party.identifier', 'party', "Identifiers",
         help="Add other identifiers of the party.")
     tax_identifier = fields.Function(fields.Many2One(
             'party.identifier', 'Tax Identifier',
             help="The identifier used for tax report."),
         'get_tax_identifier', searcher='search_tax_identifier')
-    addresses = fields.One2Many('party.address', 'party',
-        'Addresses', states=STATES, depends=DEPENDS)
-    contact_mechanisms = fields.One2Many('party.contact_mechanism', 'party',
-        'Contact Mechanisms', states=STATES, depends=DEPENDS)
-    categories = fields.Many2Many('party.party-party.category',
-        'party', 'category', 'Categories', states=STATES, depends=DEPENDS,
+    addresses = fields.One2Many('party.address', 'party', "Addresses")
+    contact_mechanisms = fields.One2Many(
+        'party.contact_mechanism', 'party', "Contact Mechanisms")
+    categories = fields.Many2Many(
+        'party.party-party.category', 'party', 'category', "Categories",
         help="The categories the party belongs to.")
     replaced_by = fields.Many2One('party.party', "Replaced By", readonly=True,
         states={
@@ -70,16 +64,17 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     fax = fields.Function(fields.Char('Fax'), 'get_mechanism')
     email = fields.Function(fields.Char('E-Mail'), 'get_mechanism')
     website = fields.Function(fields.Char('Website'), 'get_mechanism')
+    distance = fields.Function(fields.Integer('Distance'), 'get_distance')
 
     @classmethod
     def __setup__(cls):
         super(Party, cls).__setup__()
         t = cls.__table__()
         cls._sql_constraints = [
-            ('code_uniq', Unique(t, t.code),
-             'The code of the party must be unique.')
-        ]
-        cls._order.insert(0, ('name', 'ASC'))
+            ('code_uniq', Unique(t, t.code), 'party.msg_party_code_unique')
+            ]
+        cls._order.insert(0, ('distance', 'ASC NULLS LAST'))
+        cls._order.insert(1, ('name', 'ASC'))
         cls.active.states.update({
                 'readonly': Bool(Eval('replaced_by')),
                 })
@@ -126,11 +121,22 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
         return True
 
     @classmethod
-    def _tax_identifier_types(cls):
-        return ['eu_vat']
+    def tax_identifier_types(cls):
+        return [
+            'ad_nrt', 'al_nipt', 'ar_cuit', 'be_vat', 'bg_vat', 'by_unp',
+            'ch_vat', 'cl_rut', 'cn_uscc', 'co_rut', 'cr_cpj', 'cz_dic',
+            'de_vat', 'dk_cvr', 'do_rnc', 'ec_ruc', 'ee_kmkr', 'es_nif',
+            'eu_vat', 'fi_alv', 'fr_tva', 'gb_vat', 'gr_vat', 'gt_nit',
+            'hu_anum', 'id_npwp', 'ie_vat', 'il_hp', 'is_vsk', 'it_iva',
+            'jp_cn', 'kr_brn', 'lt_pvm', 'lu_tva', 'lv_pvn', 'mc_tva',
+            'md_idno', 'mt_vat', 'mx_rfc', 'nl_btw', 'no_mva', 'nz_ird',
+            'pe_ruc', 'pl_nip', 'pt_nif', 'py_ruc', 'ro_cf', 'rs_pib',
+            'ru_inn', 'se_vat', 'si_ddv', 'sk_dph', 'sm_coe', 'us_atin',
+            'us_ein', 'us_itin', 'us_ptin', 'us_ssn', 'us_tin', 'uy_ruc',
+            've_rif', 'za_tin']
 
     def get_tax_identifier(self, name):
-        types = self._tax_identifier_types()
+        types = self.tax_identifier_types()
         for identifier in self.identifiers:
             if identifier.type in types:
                 return identifier.id
@@ -138,7 +144,7 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
     @classmethod
     def search_tax_identifier(cls, name, clause):
         _, operator, value = clause
-        types = cls._tax_identifier_types()
+        types = cls.tax_identifier_types()
         domain = [
             ('identifiers', 'where', [
                     ('code', operator, value),
@@ -166,15 +172,63 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
                 return mechanism.value
         return ''
 
+    @classmethod
+    def _distance_query(cls, usages=None, party=None, depth=None):
+        context = Transaction().context
+        if party is None:
+            party = context.get('related_party')
+
+        if not party:
+            return
+
+        table = cls.__table__()
+        return table.select(
+            table.id.as_('to'),
+            Literal(0).as_('distance'),
+            where=(table.id == party))
+
+    @classmethod
+    def get_distance(cls, parties, name):
+        distances = {p.id: None for p in parties}
+        query = cls._distance_query()
+        if query:
+            cursor = Transaction().connection.cursor()
+            cursor.execute(*query.select(
+                    query.to.as_('to'),
+                    Min(query.distance).as_('distance'),
+                    group_by=[query.to]))
+            distances.update(cursor)
+        return distances
+
+    @classmethod
+    def order_distance(cls, tables):
+        party, _ = tables[None]
+        key = 'distance'
+        if key not in tables:
+            query = cls._distance_query()
+            if not query:
+                return []
+            query = query.select(
+                    query.to.as_('to'),
+                    Min(query.distance).as_('distance'),
+                    group_by=[query.to])
+            join = party.join(query, type_='LEFT',
+                condition=query.to == party.id)
+            tables[key] = {
+                None: (join.right, join.condition),
+                }
+        else:
+            query, _ = tables[key][None]
+        return [query.distance]
+
     @classmethod
     def _new_code(cls, **pattern):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('party.configuration')
         config = Configuration(1)
         sequence = config.get_multivalue('party_sequence', **pattern)
         if sequence:
-            return Sequence.get_id(sequence.id)
+            return sequence.get()
 
     @classmethod
     def create(cls, vlist):
@@ -211,9 +265,12 @@ class Party(DeactivableMixin, ModelSQL, ModelView, MultiValueMixin):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
+        code_value = clause[2]
+        if clause[1].endswith('like'):
+            code_value = lstrip_wildcard(clause[2])
         return [bool_op,
-            ('code',) + tuple(clause[1:]),
-            ('identifiers.code',) + tuple(clause[1:]),
+            ('code', clause[1], code_value) + tuple(clause[3:]),
+            ('identifiers.code', clause[1], code_value) + tuple(clause[3:]),
             ('name',) + tuple(clause[1:]),
             ('contact_mechanisms.rec_name',) + tuple(clause[1:]),
             ]
@@ -265,9 +322,8 @@ class PartyLang(ModelSQL, ValueMixin):
     def __register__(cls, module_name):
         pool = Pool()
         Party = pool.get('party.party')
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
         table = cls.__table__()
         party = Party.__table__()
 
@@ -303,36 +359,167 @@ class PartyCategory(ModelSQL):
         ondelete='CASCADE', required=True, select=True)
 
 
-class PartyIdentifier(sequence_ordered(), ModelSQL, ModelView):
+IDENTIFIER_TYPES = [
+    ('ad_nrt', "Andorra Tax Number"),
+    ('al_nipt', "Albanian VAT Number"),
+    ('ar_cuit', "Argentinian Tax Number"),
+    ('ar_dni', "Argentinian National Identity Number"),
+    ('at_businessid', "Austrian Company Register"),
+    ('at_tin', "Austrian Tax Identification"),
+    ('at_vnr', "Austrian Social Security Number"),
+    ('au_abn', "Australian Business Number"),
+    ('au_acn', "Australian Company Number"),
+    ('au_tfn', "Australian Tax File Number"),
+    ('be_vat', "Belgian Enterprise Number"),
+    ('bg_egn', "Bulgarian Personal Identity Codes"),
+    ('bg_pnf', "Bulgarian Number of a Foreigner"),
+    ('bg_vat', "Bulgarian VAT Number"),
+    ('br_cnpj', "Brazillian Company Identifier"),
+    ('br_cpf', "Brazillian National Identifier"),
+    ('by_unp', "Belarus VAT Number"),
+    ('ca_bn', "Canadian Business Number"),
+    ('ca_sin', "Canadian Social Insurance Number"),
+    ('ch_ssn', "Swiss Social Security Number"),
+    ('ch_uid', "Swiss Business Identifier"),
+    ('ch_vat', "Swiss VAT Number"),
+    ('cl_rut', "Chilean National Tax Number"),
+    ('cn_ric', "Chinese Resident Identity Card Number"),
+    ('cn_uscc', "Chinese Unified Social Credit Code"),
+    ('co_nit', "Colombian Identity Code"),
+    ('co_rut', "Colombian Business Tax Number"),
+    ('cr_cpf', "Costa Rica Physical Person ID Number"),
+    ('cr_cpj', "Costa Rica Tax Number"),
+    ('cr_cr', "Costa Rica Foreigners ID Number"),
+    ('cu_ni', "Cuban Identity Card Number"),
+    ('cy_vat', "Cypriot VAT Number"),
+    ('cz_dic', "Czech VAT Number"),
+    ('cz_rc', "Czech National Identifier"),
+    ('de_handelsregisternummer', "German Company Register Number"),
+    ('de_idnr', "German Personal Tax Number"),
+    ('de_stnr', "German Tax Number"),
+    ('de_vat', "German VAT Number"),
+    ('dk_cpr', "Danish Citizen Number"),
+    ('dk_cvr', "Danish VAT Number"),
+    ('do_cedula', "Dominican Republic National Identification Number"),
+    ('do_rnc', "Dominican Republic Tax"),
+    ('ec_ci', "Ecuadorian Personal Identity Code"),
+    ('ec_ruc', "Ecuadorian Tax Identification"),
+    ('ee_ik', "Estonian Personal ID Number"),
+    ('ee_kmkr', "Estonian VAT Number"),
+    ('ee_registrikood', "Estonian Organisation Registration Code"),
+    ('es_cif', "Spanish Company Tax"),
+    ('es_dni', "Spanish Personal Identity Codes"),
+    ('es_nie', "Spanish Foreigner Number"),
+    ('es_nif', "Spanish VAT Number"),
+    ('eu_at_02', "SEPA Identifier of the Creditor (AT-02)"),
+    ('eu_vat', "European VAT Number"),
+    ('fi_alv', "Finnish VAT Number"),
+    ('fi_associationid', "Finnish Association Identifier"),
+    ('fi_hetu', "Finnish Personal Identity Code"),
+    ('fi_veronumero', "Finnish Individual Tax Number"),
+    ('fi_ytunnus', "Finnish Business Identifier"),
+    ('fr_nif', "French Tax Identification Number"),
+    ('fr_nir', "French Personal Identification Number"),
+    # TODO: remove from party_siren
+    # ('fr_siren', "French Company Identification Number"),
+    ('fr_tva', "French VAT Number"),
+    ('gb_nhs',
+        "United Kingdom National Health Service Patient Identifier"),
+    ('gb_upn', "English Unique Pupil Number"),
+    ('gb_vat', "United Kingdom (and Isle of Man) VAT Number"),
+    ('gr_amka', "Greek Social Security Number"),
+    ('gr_vat', "Greek VAT Number"),
+    ('gt_nit', "Guatemala Tax Number"),
+    ('hr_oib', "Croatian Identification Number"),
+    ('hu_anum', "Hungarian VAT Number"),
+    ('id_npwp', "Indonesian VAT Number"),
+    ('ie_pps', "Irish Personal Number"),
+    ('ie_vat', "Irish VAT Number"),
+    ('il_hp', "Israeli Company Number"),
+    ('il_idnr', "Israeli Identity Number"),
+    ('in_aadhaar', "Indian Digital Resident Personal Identity Number"),
+    ('in_pan', "Indian Income Tax Identifier"),
+    ('is_kennitala',
+        "Icelandic Personal and Organisation Identity Code"),
+    ('is_vsk', "Icelandic VAT Number"),
+    ('it_codicefiscale', "Italian Tax Code for Individuals"),
+    ('it_iva', "Italian VAT Number"),
+    ('jp_cn', "Japanese Corporate Number"),
+    ('kr_brn', "South Korea Business Registration Number"),
+    ('kr_krn', "South Korean Resident Registration Number"),
+    ('lt_asmens', "Lithuanian Personal Number"),
+    ('lt_pvm', "Lithuanian VAT Number"),
+    ('lu_tva', "Luxembourgian VAT Number"),
+    ('lv_pvn', "Latvian VAT Number"),
+    ('mc_tva', "Monacan VAT Number"),
+    ('md_idno', "Moldavian Company Identification Number"),
+    ('mt_vat', "Maltese VAT Number"),
+    ('mu_nid', "Mauritian National Identifier"),
+    ('mx_rfc', "Mexican Tax Number"),
+    ('my_nric',
+        "Malaysian National Registration Identity Card Number"),
+    ('nl_brin', "Dutch School Identification Number"),
+    ('nl_bsn', "Dutch Citizen Identification Number"),
+    ('nl_btw', "Dutch VAT Number"),
+    ('nl_onderwijsnummer', "Dutch Student Identification Number"),
+    ('no_fodselsnummer',
+        "Norwegian Birth Number, the National Identity Number"),
+    ('no_mva', "Norwegian VAT Number"),
+    ('no_orgnr', "Norwegian Organisation Number"),
+    ('nz_ird', "New Zealand Inland Revenue Department Number"),
+    ('pe_cui', "Peruvian Identity Number"),
+    ('pe_ruc', "Peruvian Company Tax Number"),
+    ('pl_nip', "Polish VAT Number"),
+    ('pl_pesel', "Polish National Identification Number"),
+    ('pl_regon', "Polish Register of Economic Units"),
+    ('pt_nif', "Portuguese VAT Number"),
+    ('py_ruc', "Paraguay Tax Number"),
+    ('ro_cf', "Romanian VAT Number"),
+    ('ro_cnp', "Romanian Numerical Personal Code"),
+    ('ro_onrc', "Romanian ONRC Number"),
+    ('rs_pib', "Serbian Tax Identification"),
+    ('ru_inn', "Russian Tax identifier"),
+    ('se_orgnr', "Swedish Company Number"),
+    ('se_personnummer', "Swedish Personal Number"),
+    ('se_vat', "Swedish VAT Number"),
+    ('si_ddv', "Slovenian VAT Number"),
+    ('sk_dph', "Slovak VAT Number"),
+    ('sk_rc', "Slovak Birth Number"),
+    ('sm_coe', "San Marino National Tax Number"),
+    ('tr_tckimlik', "Turkish Personal Identification Number"),
+    ('us_atin', "U.S. Adoption Taxpayer Identification Number"),
+    ('us_ein', "U.S. Employer Identification Number"),
+    ('us_itin', "U.S. Individual Taxpayer Identification Number"),
+    ('us_ptin', "U.S. Preparer Tax Identification Number"),
+    ('us_ssn', "U.S. Social Security Number"),
+    ('us_tin', "U.S. Taxpayer Identification Number"),
+    ('uy_ruc', "Uruguay Tax Number"),
+    ('ve_rif', "Venezuelan VAT Number"),
+    ('za_idnr', "South African Identity Document Number"),
+    ('za_tin', "South African Tax Identification Number"),
+    ]
+
+
+class Identifier(sequence_ordered(), ModelSQL, ModelView):
     'Party Identifier'
     __name__ = 'party.identifier'
     _rec_name = 'code'
     party = fields.Many2One('party.party', 'Party', ondelete='CASCADE',
         required=True, select=True,
         help="The party identified by this record.")
-    type = fields.Selection([
-            (None, ''),
-            ('eu_vat', 'VAT'),
-            ], 'Type')
+    type = fields.Selection('get_types', 'Type')
     type_string = type.translated('type')
     code = fields.Char('Code', required=True)
 
-    @classmethod
-    def __setup__(cls):
-        super(PartyIdentifier, cls).__setup__()
-        cls._error_messages.update({
-                'invalid_vat': ('Invalid VAT number "%(code)s" '
-                    'on party "%(party)s".'),
-                })
-
     @classmethod
     def __register__(cls, module_name):
         pool = Pool()
         Party = pool.get('party.party')
         cursor = Transaction().connection.cursor()
         party = Party.__table__()
+        table = cls.__table__()
 
-        super(PartyIdentifier, cls).__register__(module_name)
+        super().__register__(module_name)
 
         party_h = Party.__table_handler__(module_name)
         if (party_h.column_exist('vat_number')
@@ -342,44 +529,63 @@ class PartyIdentifier(sequence_ordered(), ModelSQL, ModelView):
                     party.id, party.vat_number, party.vat_country,
                     where=(party.vat_number != Null)
                     | (party.vat_country != Null)))
-            for party_id, number, country in cursor.fetchall():
+            for party_id, number, country in cursor:
                 code = (country or '') + (number or '')
                 if not code:
                     continue
-                type = None
-                if vat.is_valid(code):
-                    type = 'eu_vat'
+                for type in Party.tax_identifier_types():
+                    module = get_cc_module(*type.split('_', 1))
+                    if module.is_valid(code):
+                        break
+                else:
+                    type = None
                 identifiers.append(
                     cls(party=party_id, code=code, type=type))
             cls.save(identifiers)
             party_h.drop_column('vat_number')
             party_h.drop_column('vat_country')
 
+        # Migration from 5.8: Rename cn_rit into cn_ric
+        cursor.execute(*table.update([table.type], ['cn_ric'],
+                where=(table.type == 'cn_rit')))
+
+    @classmethod
+    def get_types(cls):
+        pool = Pool()
+        Configuration = pool.get('party.configuration')
+        configuration = Configuration(1)
+        return [(None, '')] + configuration.get_identifier_types()
+
     @fields.depends('type', 'code')
     def on_change_with_code(self):
-        if self.type == 'eu_vat':
-            try:
-                return vat.compact(self.code)
-            except stdnum.exceptions.ValidationError:
-                pass
+        if self.type and '_' in self.type:
+            module = get_cc_module(*self.type.split('_', 1))
+            if module:
+                try:
+                    return module.compact(self.code)
+                except stdnum.exceptions.ValidationError:
+                    pass
         return self.code
 
     def pre_validate(self):
-        super(PartyIdentifier, self).pre_validate()
+        super().pre_validate()
         self.check_code()
 
     @fields.depends('type', 'party', 'code')
     def check_code(self):
-        if self.type == 'eu_vat':
-            if not vat.is_valid(self.code):
-                if self.party and self.party.id > 0:
-                    party = self.party.rec_name
-                else:
-                    party = ''
-                self.raise_user_error('invalid_vat', {
-                        'code': self.code,
-                        'party': party,
-                        })
+        if self.type and '_' in self.type:
+            module = get_cc_module(*self.type.split('_', 1))
+            if module:
+                if not module.is_valid(self.code):
+                    if self.party and self.party.id > 0:
+                        party = self.party.rec_name
+                    else:
+                        party = ''
+                    raise InvalidIdentifierCode(
+                        gettext('party.msg_invalid_code',
+                            type=self.type_string,
+                            code=self.code,
+                            party=party))
 
 
 class CheckVIESResult(ModelView):
@@ -406,26 +612,16 @@ class CheckVIES(Wizard):
             Button('OK', 'end', 'tryton-ok', True),
             ])
 
-    @classmethod
-    def __setup__(cls):
-        super(CheckVIES, cls).__setup__()
-        cls._error_messages.update({
-                'vies_unavailable': ('The VIES service is unavailable, '
-                    'try again later.'),
-                })
-
     def transition_check(self):
-        Party = Pool().get('party.party')
-
         parties_succeed = []
         parties_failed = []
-        parties = Party.browse(Transaction().context.get('active_ids'))
-        for party in parties:
+        for party in self.records:
             for identifier in party.identifiers:
                 if identifier.type != 'eu_vat':
                     continue
+                eu_vat = get_cc_module('eu', 'vat')
                 try:
-                    if not vat.check_vies(identifier.code):
+                    if not eu_vat.check_vies(identifier.code)['valid']:
                         parties_failed.append(party.id)
                     else:
                         parties_succeed.append(party.id)
@@ -439,7 +635,8 @@ class CheckVIES(Wizard):
                                 or e.faultstring.find('MS_UNAVAILABLE') \
                                 or e.faultstring.find('TIMEOUT') \
                                 or e.faultstring.find('SERVER_BUSY'):
-                            self.raise_user_error('vies_unavailable')
+                            raise VIESUnavailable(
+                                gettext('party.msg_vies_unavailable')) from e
                     raise
         self.result.parties_succeed = parties_succeed
         self.result.parties_failed = parties_failed
@@ -452,7 +649,7 @@ class CheckVIES(Wizard):
             }
 
 
-class PartyReplace(Wizard):
+class Replace(Wizard):
     "Replace Party"
     __name__ = 'party.replace'
     start_state = 'ask'
@@ -462,26 +659,19 @@ class PartyReplace(Wizard):
             ])
     replace = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(PartyReplace, cls).__setup__()
-        cls._error_messages.update({
-                'different_name': ("Parties have different names: "
-                    "%(source_name)s vs %(destination_name)s."),
-                'different_tax_identifier': (
-                    "Parties have different Tax Identifier: "
-                    "%(source_code)s vs %(destination_code)s."),
-                })
-
     def check_similarity(self):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
         source = self.ask.source
         destination = self.ask.destination
         if source.name != destination.name:
             key = 'party.replace name %s %s' % (source.id, destination.id)
-            self.raise_user_warning(key, 'different_name', {
-                    'source_name': source.name,
-                    'destination_name': destination.name,
-                    })
+            if Warning.check(key):
+                raise SimilarityWarning(
+                    key,
+                    gettext('party.msg_different_name',
+                        source_name=source.name,
+                        destination_name=destination.name))
         source_code = (source.tax_identifier.code
             if source.tax_identifier else '')
         destination_code = (destination.tax_identifier.code
@@ -489,10 +679,12 @@ class PartyReplace(Wizard):
         if source_code != destination_code:
             key = 'party.replace tax_identifier %s %s' % (
                 source.id, destination.id)
-            self.raise_user_warning(key, 'different_tax_identifier', {
-                    'source_code': source_code,
-                    'destination_code': destination_code,
-                    })
+            if Warning.check(key):
+                raise SimilarityWarning(
+                    key,
+                    gettext('party.msg_different_tax_identifier',
+                        source_code=source_code,
+                        destination_code=destination_code))
 
     def transition_replace(self):
         pool = Pool()
@@ -548,7 +740,7 @@ class PartyReplace(Wizard):
             ]
 
 
-class PartyReplaceAsk(ModelView):
+class ReplaceAsk(ModelView):
     "Replace Party"
     __name__ = 'party.replace.ask'
     source = fields.Many2One('party.party', "Source", required=True,
@@ -572,7 +764,7 @@ class PartyReplaceAsk(ModelView):
             self.destination = self.source.replaced_by
 
 
-class PartyErase(Wizard):
+class Erase(Wizard):
     "Erase Party"
     __name__ = 'party.erase'
     start_state = 'ask'
@@ -582,15 +774,6 @@ class PartyErase(Wizard):
             ])
     erase = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(PartyErase, cls).__setup__()
-        cls._error_messages.update({
-                'active_party': (
-                    'The party "%(party)s" can not be erased '
-                    'because he is still active.'),
-                })
-
     def transition_erase(self):
         pool = Pool()
         Party = pool.get('party.party')
@@ -656,15 +839,14 @@ class PartyErase(Wizard):
                                         Model.__name__ + ',%')
                                     & Model.id.sql_cast(
                                         Substring(table.resource,
-                                            Position(',', table.resource) +
-                                            Literal(1))).in_(query)))
+                                            Position(',', table.resource)
+                                            + Literal(1))).in_(query)))
         return 'end'
 
     def check_erase(self, party):
         if party.active:
-            self.raise_user_error('active_party', {
-                    'party': party.rec_name,
-                    })
+            raise EraseError(gettext('party.msg_erase_active_party',
+                    party=party.rec_name))
 
     def to_erase(self, party_id):
         pool = Pool()
@@ -680,7 +862,8 @@ class PartyErase(Wizard):
                 ['type', 'code'],
                 [None, '****']),
             (Address, [('party', '=', party_id)], True,
-                ['name', 'street', 'zip', 'city', 'country', 'subdivision'],
+                ['name', 'street', 'postal_code', 'city',
+                    'country', 'subdivision'],
                 [None, None, None, None, None, None]),
             (ContactMechanism, [('party', '=', party_id)], True,
                 ['value', 'name', 'comment'],
@@ -692,10 +875,11 @@ class PartyErase(Wizard):
         pool = Pool()
         Attachment = pool.get('ir.attachment')
         Note = pool.get('ir.note')
-        return [Attachment, Note]
+        Avatar = pool.get('ir.avatar')
+        return [Attachment, Note, Avatar]
 
 
-class PartyEraseAsk(ModelView):
+class EraseAsk(ModelView):
     "Erase Party"
     __name__ = 'party.erase.ask'
     party = fields.Many2One('party.party', "Party", required=True,
